import base64
import errno
import http.client
import os
import simplejson
import ssl
import time

from cmd import kubectl_command
from constants import PRTON_PORT
from constants import GET_AAA_ROLE_BINDING
from constants import GET_CLUSTER_CERT
from constants import GET_CLUSTER_CERT_WITH_PRIVATE_KEY
from constants import PRIVATE_KEY_PATH
from constants import PUBLIC_CERT_PATH
from constants import NAPP_PLATFORM_CA_PATH
from logger import get_logger

_SSL_CONTEXT = None
_CERT_MANAGER_NS = "cert-manager"

lg = get_logger(__name__)


def assert_request_success(ret_and_err, description="API request"):
    ret = ret_and_err[0]
    error_msg = ret_and_err[1]
    if ret is None:
        lg.error("Description: %s", description)
        lg.error("Request failed with error msg: %s" % error_msg)
        raise RuntimeError("Request failed with error msg: %s" % error_msg)
    return ret


def remote_node_request_with_retry(host, https_method, req_uri, body=None,
                                   headers=None, max_retry=5, interval=10):
    ret, err = remote_node_request(host, https_method, req_uri, body, headers)
    count = 0
    interval = interval if interval > 0 else 1
    max_retry = max_retry if max_retry > 0 else 0
    while ret is None and count < max_retry:
        time.sleep(interval)
        count = count + 1
        ret, err = remote_node_request(host, https_method, req_uri, body,
                                       headers)
        if ret is None:
            lg.warn("Retry #%s: Remote node request failed with error msg: %s,"
                    % (count, err))
        else:
            return ret, err
    return ret, err


def remote_node_request(host, https_method, req_uri, body=None,
                        headers=None):
    """
    Issue REST request again remote API service.

    :param host: the remote host on which webserver is running
    :param https_method: GET, PUT, POST
    :param req_uri: URI
    :param body: body: optional body (e.g. python dictionary) to issue.
    :param headers: optional headers to issue (application/json is
            automatically added if body is not None).
    :return:
        (resp_body, error_msg)
        resp_body: the response body as a json object. None if request fails
        error_msg: the error msg if request fails. None if request successes.
    """
    headers = headers or {}
    if "X-NSX-Username" not in headers:
        headers["X-NSX-Username"] = _get_admin_username()
    if "x-nsx-roles" not in headers:
        headers['x-nsx-roles'] = "ZW50ZXJwcmlzZV9hZG1pbgo="
    context = _get_ssl_context()
    conn = http.client.HTTPSConnection(host, port="443", context=context)
    return _rest_request(conn, https_method, req_uri, body, headers)


def local_node_request(port, http_method, req_uri, body=None, headers=None):
    """
    Issue REST request against local API service.

    :param port: the port on which the webserver is running on localhost
    :param http_method: GET, PUT, POST, ...
    :param req_uri: URI
    :param body: optional body (e.g. python dictionary) to issue.
    :param headers: optional headers to issue (application/json is
            automatically added if body is not None).
    :return:
        (resp_body, error_msg)
        resp_body: the response body as a json object. None if request fails
        error_msg: the error msg if request fails. None if request successes.
    """
    headers = headers or {}
    if "X-NSX-Username" not in headers:
        headers["X-NSX-Username"] = _get_admin_username()
    port = port or PRTON_PORT
    host_and_port = "127.0.0.1:%s" % port
    conn = http.client.HTTPConnection(host_and_port)
    return _rest_request(conn, http_method, req_uri, body, headers)


def _get_admin_username():
    default_admin_username = "admin"

    host_and_port = "127.0.0.1:%s" % PRTON_PORT
    conn = http.client.HTTPConnection(host_and_port)
    resp, err = _rest_request(conn, "GET", GET_AAA_ROLE_BINDING)
    if resp is None:
        lg.warn("Request to fetch admin user-name failed with error msg: %s", err)
        return default_admin_username

    try:
        for result in resp['results']:
            if result['type'] != "local_user":
                continue
            for role in result['roles']:
                if role['role'] == "enterprise_admin":
                    admin_username = result['name']
                    lg.info("Found admin user-name: %s", admin_username)
                    return admin_username
    except Exception as ex:
        lg.exception("Exception encountered while fetching admin user-name: %s", str(ex))

    lg.warn("Using default admin user-name: %s", default_admin_username)
    return default_admin_username


def _rest_request(conn, http_method, req_uri, body=None, headers=None):
    headers = headers or {}
    json_body = None
    # make sure 'body' is a json object
    if body is not None:
        try:
            json_body = simplejson.dumps(body)
        except Exception as ex:
            lg.exception("Invalid request: body is not a json object")
            return None, "Invalid request: body is not a json object"
        headers["Content-Type"] = "application/json"

    try:
        lg.info("%s:%s", http_method, req_uri)
        conn.request(http_method, req_uri, json_body, headers)
        resp = conn.getresponse()
        resp_body = resp.read()
        lg.info(resp_body)
        if 200 <= resp.status <= 299:
            if not resp_body:
                return {}, None
            try:
                resp_obj = simplejson.loads(resp_body)
                return resp_obj, None
            except TypeError as json_ex:
                lg.exception("Error converting json response body for %s %s",
                             http_method, req_uri)
                return None, "Error loading response JSON: %s" % str(json_ex)
        else:
            lg.error("Unexpected error for %s %s, status: %s, body: %s",
                     http_method, req_uri, resp.status, resp_body)
            return (None, "%s %s returned status: %s, body: %s" %
                    (http_method, req_uri, resp.status, resp_body))
    except Exception as ex:
        if hasattr(ex, "errno") and ex.errno == errno.ECONNREFUSED:
            lg.exception("%s: %s:%s", "Connection refused", conn.host,
                         conn.port)
            return None, "%s: %s:%s" % (
                "Connection refused", conn.host, conn.port)
        else:
            lg.exception("Error invoking manager request %s %s", http_method,
                         req_uri)
            return None, "Exception: %s" % str(ex)


def _get_ssl_context():
    global _SSL_CONTEXT
    if _SSL_CONTEXT is not None:
        return _SSL_CONTEXT
    _prepare_ca_cert()
    _prepare_key_pair()
    context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    context.verify_mode = ssl.CERT_REQUIRED
    context.load_verify_locations(NAPP_PLATFORM_CA_PATH)
    context.load_cert_chain(PUBLIC_CERT_PATH, PRIVATE_KEY_PATH)
    _SSL_CONTEXT = context
    return _SSL_CONTEXT


def _prepare_ca_cert():
    command = r"get secret -n " + _CERT_MANAGER_NS + " ca-key-pair " \
                                                     r"-ojsonpath='{" \
                                                     r".data.tls\.crt}' "
    rc, out, error = kubectl_command(command)
    if rc is not 0:
        lg.error("Failed to get ca cert with error msg: %s" % error)
        raise RuntimeError("Failed to get ca cert with error msg: %s" % error)
    with open(NAPP_PLATFORM_CA_PATH, "w+") as w:
        w.write(base64.b64decode(out).decode('utf-8'))
    os.chmod(NAPP_PLATFORM_CA_PATH, 600)


def _prepare_key_pair():
    ret, error_msg = local_node_request(PRTON_PORT, "GET", GET_CLUSTER_CERT)
    if ret is None:
        lg.error("Failed to get key pair with error msg: %s" % error_msg)
        raise RuntimeError("Failed to get key pair with error msg: %s" %
                           error_msg)
    lg.info(ret)
    cluster_cert_id = ret['results'][0]['id']

    # get the private key and public cert
    uri = GET_CLUSTER_CERT_WITH_PRIVATE_KEY.replace("<id>", cluster_cert_id)
    ret, error_msg = local_node_request(PRTON_PORT, "GET", uri)
    if ret is None:
        lg.error("Failed to get key pair with error msg: %s" % error_msg)
        raise RuntimeError("Failed to get key pair with error msg: %s" %
                           error_msg)

    pem_encoded = ret['pem_encoded']
    private_key = ret['private_key']
    with open(PRIVATE_KEY_PATH, "w+") as w:
        w.write(private_key)
    os.chmod(PRIVATE_KEY_PATH, 600)
    with open(PUBLIC_CERT_PATH, "w+") as w:
        w.write(pem_encoded)
    os.chmod(PUBLIC_CERT_PATH, 600)


def set_cert_manager_ns(ns):
    if ns:
        global _CERT_MANAGER_NS
        _CERT_MANAGER_NS = ns
