#!/usr/bin/env python3
import argparse
import os
import shutil
import sys
import time
from pathlib import Path

from api_request import assert_request_success
from api_request import local_node_request
from api_request import remote_node_request_with_retry
from api_request import set_cert_manager_ns
from cmd import kubectl_command
from cmd import set_kubectl
from constants import Certificate
from constants import DELETE_CERT
from constants import DELETE_CSR
from constants import DELETE_PI
from constants import GET_ALL_CERTS
from constants import GET_CERT_ID
from constants import GET_CLOUDNATIVE_PLATFORM_CERTS
from constants import GET_DEPLOYMENT_REGISTRY
from constants import GET_FEATURES_INTELLIGENCE_CONFIG
from constants import GET_NSX_NODE_VERSION
from constants import GET_PI_CERTS
from constants import GET_PLATFORM_HEALTH
from constants import HELM_CA_41X_WORKAROUND_FILENAME
from constants import NAPI_PORT
from constants import NAPP_HELM_CA_ALIAS
from constants import NAPP_CERT_EGRESS_ENTITIES
from constants import NAPP_CERT_NONEGRESS_ENTITIES
from constants import POST_APPLY_KAFKA_CLIENT_CERT
from constants import POST_CLOUDNATIVE_PLATFROM_CERT
from constants import POST_GENERATE_CSR
from constants import POST_REGISTER_PI_CERT
from constants import POST_SELF_SIGN_CSR
from constants import PRTON_PORT
from constants import RELEASE_USED_CERTIFICATE
from constants import TIMEOUT_SECS_LOAD_SITE
from constants import USED_BY_NAPP_HELM_CA
from constants import WORK_DIR
from logger import get_logger
from logger import setup_logger

lg = get_logger(__name__)


def _get_ingress_fqdn(namespace: str) -> str:
    command = r"get httpproxy cloudnative-platform-ingress -n %s " \
              r"-ojsonpath='{.spec.virtualhost.fqdn}' " % namespace
    rc, out, error = kubectl_command(command)
    if rc != 0:
        lg.error("Failed to get ingress FQDN with error msg: %s" % error)
        raise RuntimeError("Failed to get FQDN with error msg: %s" % error)
    lg.info("Ingress FQDN: %s" % out)
    return out.replace("'", "")


def _wait_util_platform_ready(host: str, max_retry: int = 45,
                              interval: float = 20):
    attempt = 0
    while attempt < max_retry:
        ret, _ = remote_node_request_with_retry(host, "GET",
                                                GET_PLATFORM_HEALTH,
                                                max_retry=1)
        if ret is not None and "health" in ret and ret['health'] == "UP":
            lg.info("All Platform services are running")
            return
        attempt = attempt + 1
        time.sleep(interval)

    lg.error("Platform services are unhealthy")
    raise RuntimeError("Platform services are unhealthy")


# this is needed for H -> I migration
def _get_features_intelligence_config(host: str):
    lg.info("Get Feature Intelligence Config")
    assert_request_success(
        remote_node_request_with_retry(host, "GET",
                                       GET_FEATURES_INTELLIGENCE_CONFIG,
                                       max_retry=5),
        "GET: %s" % GET_FEATURES_INTELLIGENCE_CONFIG)


# Create a private key and self-signed certificate for common agent/pace agent.
# It's used as keystore
def _register_kafka_client_cert():
    body = {
        "display_name": Certificate.display_name,
        "algorithm": Certificate.algorithm,
        "key_size": Certificate.key_size,
        "subject": Certificate.subject
    }
    # 1. generate a csr
    ret = assert_request_success(local_node_request(PRTON_PORT, "POST",
                                                    POST_GENERATE_CSR, body),
                                 "POST: %s" % POST_GENERATE_CSR)
    csr_id = ret['id']
    # 2. self signed the csr
    ret = assert_request_success(
        local_node_request(PRTON_PORT, "POST",
                           POST_SELF_SIGN_CSR.replace("<id>", csr_id), None),
        "POST: %s" % POST_SELF_SIGN_CSR)
    cert = ret['id']
    # 3. apply the cert
    assert_request_success(
        local_node_request(PRTON_PORT, "POST",
                           POST_APPLY_KAFKA_CLIENT_CERT.replace("<id>", cert),
                           None),
        "POST: %s" % POST_APPLY_KAFKA_CLIENT_CERT)
    # 4. delete the csr
    assert_request_success(
        local_node_request(PRTON_PORT, "DELETE",
                           DELETE_CSR.replace("<id>", csr_id), None),
        "DELETE: %s" % DELETE_CSR)
    lg.info("kafka client cert is generated and imported")


# release K8S_MSG_CLIENT certificate on manager
def _unregister_kafka_client_cert():
    # in case the k8s_msg_client has been attached by other role's certificate
    headers = {'X-Allow-Overwrite': True}
    ret = assert_request_success(
        local_node_request(PRTON_PORT, "GET", GET_ALL_CERTS, None, headers),
        "GET: %s" % GET_ALL_CERTS)
    # release K8S_MSG_CLIENT certificate on manager
    for cert in ret['results']:
        used_by = cert['used_by']
        if used_by:
            for ele in used_by:
                if Certificate.service_type in ele['service_types']:
                    body = {
                        "node_id": ele['node_id'],
                        "service_type": Certificate.service_type
                    }
                    cert_id = cert['id']
                    # release certificate
                    assert_request_success(
                        local_node_request(PRTON_PORT, "POST",
                                           RELEASE_USED_CERTIFICATE.
                                           replace("<id>", cert_id),
                                           body=body),
                        "POST: %s" % RELEASE_USED_CERTIFICATE)

    # delete system-generated certificate used as K8S_MSG_CLIENT
    for cert in ret['results']:
        if cert['_create_user'] == 'admin' and \
                cert['resource_type'] == "certificate_self_signed" and \
                cert['display_name'] == Certificate.display_name:
            cert_id = cert['id']
            # delete certificate
            assert_request_success(
                local_node_request(PRTON_PORT, "DELETE",
                                   DELETE_CERT.
                                   replace("<id>", cert_id), None, headers),
                "DELETE: %s" % DELETE_CERT)


# Import NAPP certificates into NSX Manager.
# And create the related PIs
def _register_cloudnative_platform_certs(host: str):
    ret = assert_request_success(
        remote_node_request_with_retry(host, "GET",
                                       GET_CLOUDNATIVE_PLATFORM_CERTS, None),
        "GET: %s" % GET_CLOUDNATIVE_PLATFORM_CERTS)
    for cert in ret['results']:
        body = {}
        pi = cert['alias']
        pem = cert['pem_encoded']
        if cert['used_by'] in NAPP_CERT_EGRESS_ENTITIES:
            body = {
                "name": pi,
                "node_id": "nsx.application.platform",
                "roles_for_paths": [
                    {
                        "path": "/",
                        "roles": [
                            {
                                "role": "auditor"
                            },
                            {
                                "role": "napp_role"
                            }
                        ]
                    }
                ],
                "is_protected": "true",
                "certificate_pem": pem
            }
        elif cert['used_by'] in NAPP_CERT_NONEGRESS_ENTITIES:
            # the least privileges
            body = {
                "name": pi,
                "node_id": "nsx.application.platform",
                "role": "network_engineer",
                "is_protected": "true",
                "certificate_pem": pem
            }
        if body:
            lg.info("PI: %s", pi)
            lg.info("Cert: %s", cert)
            assert_request_success(
                local_node_request(PRTON_PORT, "POST",
                                   POST_REGISTER_PI_CERT, body),
                "POST: %s" % POST_REGISTER_PI_CERT)


# Delete NAPP PIs and related certificates from NSX Manager
def _unregister_cloudnative_platform_certs():
    # delete k8s side certificates on manager
    ret = assert_request_success(
        local_node_request(PRTON_PORT, "GET", GET_PI_CERTS),
        "GET: %s" % GET_PI_CERTS)
    # in case the PI is modified by other role, add X-Allow-Overwrite
    headers = {'X-Allow-Overwrite': True}
    for pi in ret['results']:
        if pi['node_id'] == "nsx.application.platform":
            assert_request_success(
                local_node_request(PRTON_PORT, "DELETE",
                                   DELETE_PI.replace("<id>", pi['id']), None,
                                   headers), "DELETE %s" % DELETE_PI)
            assert_request_success(
                local_node_request(PRTON_PORT, "DELETE",
                                   DELETE_CERT.replace("<id>",
                                                       pi['certificate_id']),
                                   None, headers), "DELETE: %s" % DELETE_CERT)


# Push NSX Manager certificates into NAPP
def _register_manager_certs(host: str):
    node_certs = []
    kafka_certs = []
    ret = assert_request_success(
        local_node_request(PRTON_PORT, "GET", GET_ALL_CERTS),
        "GET: %s" % GET_ALL_CERTS)
    for cert in ret['results']:
        used_by = cert['used_by']
        if used_by:
            for ele in used_by:
                if "API" in ele['service_types'] and cert['resource_type'] != "certificate_ca":
                    lg.info(ele)
                    node_certs.append(cert)
                    break
                elif "K8S_MSG_CLIENT" in ele['service_types']:
                    lg.info(ele)
                    kafka_certs.append(cert)
                    break

    lg.info("Found %s node certs on manager cluster" % str(len(node_certs)))
    lg.info("Found %s kafka client certs on manager cluster" %
            str(len(kafka_certs)))

    # 1. push manager node certs to k8s platform
    if node_certs:
        _push_certs(host, node_certs, "NSX_UA_NODE")

    # 2. push kafka client cert to k8s platform
    if kafka_certs:
        _push_certs(host, kafka_certs, "NSX_UA_KAFKA_CLIENT")

    # sleep for a while since some services will be restarted
    # Monitoring service updates platform status every 60 sec, wait until it updates
    # the latest status/unstable when kafka restarts after cert update.
    time.sleep(90)


# If we are running on 4.1, 4.0 and 3.2, check if private registry CA workaround is used
# It is safe to call both the workaround and the regular helm_ca push since
# 4.1.x deployment registry data will never include helm ca cert id
def _41x_workaround_push_helm_ca_if_used(host: str):
    ret = assert_request_success(
        local_node_request(NAPI_PORT, "GET", GET_NSX_NODE_VERSION),
        "GET: %s to check if NSX manager is running 4.1.x" % GET_NSX_NODE_VERSION)
    node_version = ret["node_version"]
    lg.info("Node is running on version '%s'" % node_version)
    if node_version.startswith("4.1.") or node_version.startswith("4.0.") or node_version.startswith("3.2."):
        # NSX manager is running 4.1.x, or before check if private registry CA workaround is applied
        if os.path.isfile(HELM_CA_41X_WORKAROUND_FILENAME):
            # file exist: HELM_CA_41X_WORKAROUND_FILENAME
            # if file found and ca content not zero:
            with open(HELM_CA_41X_WORKAROUND_FILENAME, 'r') as file:
                ca_pem = file.read()
                if not ca_pem:
                    raise RuntimeError("Failed to retrieve pem from %s on a 4.1.x setup" % HELM_CA_41X_WORKAROUND_FILENAME)
                _push_helm_ca(host, ca_pem)


# If a private helm CA is configured to deploy NAPP, push its content to NAPP
def _push_helm_ca_if_used(host: str):
    ret = assert_request_success(
        local_node_request(PRTON_PORT, "GET", GET_DEPLOYMENT_REGISTRY),
        "GET: %s to check if a private CA is used for deployment" % GET_DEPLOYMENT_REGISTRY)
    if 'certificate_id' in ret:
        cert_id = ret['certificate_id']
        # A cert id is configured for private CA
        lg.info("Deployment workflow is configure to use certificate with %s as CA" % cert_id)
        ret = assert_request_success(
            local_node_request(PRTON_PORT, "GET", GET_CERT_ID.replace("<id>", cert_id)),
            "GET: %s to retrieve private Helm CA PEM" % GET_CERT_ID.replace("<id>", cert_id))
        ca_pem = ret['pem_encoded']
        if not ca_pem:
            raise RuntimeError("Failed to retrieve pem for helm CA with certificate id of %s" % cert_id)
        _push_helm_ca(host, ca_pem)

def _push_helm_ca(host: str, ca_pem: str):
    lg.info("Private CA used for deployment has pem-encoding ending with '%s'" % ca_pem[-45:-25])
    body = {
        "alias": NAPP_HELM_CA_ALIAS,
        "pem_encoded": ca_pem,
        "used_by": USED_BY_NAPP_HELM_CA,
        "skip_sync_services": False
    }
    assert_request_success(
        remote_node_request_with_retry(host, "POST",
                                   POST_CLOUDNATIVE_PLATFROM_CERT,
                                   body=body, max_retry=3),
        "POST: %s" % POST_CLOUDNATIVE_PLATFROM_CERT)


def _push_certs(host: str, certs_list: list, used_by: str):
    for cert in certs_list:
        body = {
            "alias": cert['id'],
            "pem_encoded": cert['pem_encoded'],
            "used_by": used_by,
            "skip_sync_services": True
        }
        assert_request_success(
            remote_node_request_with_retry(host, "POST",
                                           POST_CLOUDNATIVE_PLATFROM_CERT,
                                           body=body, max_retry=3),
            "POST: %s" % POST_CLOUDNATIVE_PLATFROM_CERT)

    # restart services
    body = {
        "used_by": used_by,
        "skip_sync_services": False
    }
    assert_request_success(
        remote_node_request_with_retry(host, "POST",
                                       POST_CLOUDNATIVE_PLATFROM_CERT,
                                       body=body, max_retry=3),
        "POST: %s" % POST_CLOUDNATIVE_PLATFROM_CERT)


def _clean_up_old_certs():
    _unregister_kafka_client_cert()
    _unregister_cloudnative_platform_certs()


def _clean_up_work_dir():
    work_dir = Path(WORK_DIR)
    if work_dir.exists():
        shutil.rmtree(work_dir)


def _load_site(namespace: str):
    """
    un-suspends the load-default-site job from site-service chart. if
    it fails to un-suspend the job, raise exception and exit.

    the job does loading of nsx site information to site-service

    will wait until timeout-secs for the job to complete. if it does not
    complete within timeout, raise exception and exit.

    :param namespace: namespace which has the job resource
    :return: None
    """
    # un-suspend the job
    command = r"""patch -n %s job/load-default-site --type=strategic --patch {"spec":{"suspend":false}}""" \
              % namespace
    rc, out, error = kubectl_command(command)
    if rc != 0:
        lg.error("Failed to un-suspend load-default-site job: %s" % error)
        raise RuntimeError("Failed to un-suspend load-default-site job: %s" % error)
    # sleep for 10 secs - for spec to be applied
    time.sleep(10)
    # now wait until job is complete
    command = r"wait -n %s --for=condition=complete job/load-default-site " \
              r"--timeout %s" % (namespace, TIMEOUT_SECS_LOAD_SITE)
    rc, out, error = kubectl_command(command)
    if rc != 0:
        lg.error("Timed out waiting for load-default-site job: %s" % error)
        raise RuntimeError("Timed out waiting for load-default-site job: %s" % error)
    return


def main(args):
    fqdn = _get_ingress_fqdn(args.namespace)
    _register_cloudnative_platform_certs(fqdn)
    _register_kafka_client_cert()
    _register_manager_certs(fqdn)
    _41x_workaround_push_helm_ca_if_used(fqdn)
    _push_helm_ca_if_used(fqdn)
    _wait_util_platform_ready(fqdn)
    _load_site(args.namespace)
    _get_features_intelligence_config(fqdn)
    lg.info("Registration Successfully!!!")


if __name__ == "__main__":
    try:
        parser = argparse.ArgumentParser(description="Registration between "
                                                     "NSX Application "
                                                     "Platform and NSX "
                                                     "Manager")
        parser.add_argument('-u', "--unregister", help="cleanup NAPP "
                                                       "certificates on NSX "
                                                       "Manager",
                            required=False, action='store_true')
        parser.add_argument("-n", "--namespace", help="the namespace where "
                                                      "platform chart is "
                                                      "installed",
                            required=('-u' not in sys.argv and
                                      '--unregister' not in sys.argv))
        parser.add_argument("-c", "--cmnamespace", help="the namespace where "
                                                        "cert-manager is "
                                                        "installed",
                            default="cert-manager", required=False)
        parser.add_argument("-k", "--kubeconfig", help="Kubeconfig, the file "
                                                       "used to configure "
                                                       "access to Kubernetes "
                                                       "(Optional)",
                            default="~/.kube/config", required=False)
        args = parser.parse_args()
        # setup logger
        setup_logger()
        # old certs are always cleaned up when performing register
        _clean_up_old_certs()
        if args.unregister:
            # if action was unregister, we already achieved in previous step of _clean_up_old_certs()
            sys.exit(0)
        # cleanup workspace
        _clean_up_work_dir()
        # setup kubectl
        set_kubectl(args.kubeconfig)
        # setup cert manager namespace
        set_cert_manager_ns(args.cmnamespace)
        # setup work dir
        work_dir_path = Path(WORK_DIR)
        os.mkdir(work_dir_path)
        # main func
        main(args)
    except Exception as e:
        lg.exception("Exit unexpectedly")
        sys.exit(-1)
    finally:
        _clean_up_work_dir()
